from .exception import AuthFailure, InvalidLookupRequest, WsOpenFailure, InvalidToken, InvalidJsonPayload, AuthRefreshFailure, SubRefreshFailure
from .object_set import AciSet, AciTree, ListOfAciTree, to_tree, to_ltree
import http.client
import threading
import logging
import json
import time

logger1 = logging.getLogger(__name__)


class ApiContext:
    def __init__(self):
        self.base_header = {
                'cache-control': 'no-cache',
                'Content-Type': 'test/plain'
        }

        self.ip = ''
        self.Cookie = ''
        self.token = ''
        self.urlToken = ''
        self.subscription_list = list()
        self.lock = threading.Lock()

        self.ws_status = "CLOSE"

        self.keep_running_thread = True


    def reset_token(self):
        self.Cookie = ''
        self.token = ''
        self.urlToken = ''

    @property
    def ws_url(self):
        return "wss://{0}/socket{1}?{2}".format(self.ip, self.token, self.urlToken)

    @property
    def http_header(self):
        header = self.base_header
        header['APIC-challenge'] = self.urlToken
        header['Cookie'] = self.Cookie

        return header

    @property
    def auth_token(self):
        if self.Cookie and self.urlToken:
            return True
        else:
            return False


global api_ctx
api_ctx=ApiContext()


class HttpClient:
    def __init__(self, ip, user, password, ctx=api_ctx):
        self.__conn = http.client.HTTPSConnection(ip, http.client.HTTPS_PORT)
        # self.__conn = http.client.HTTPConnection(ip, http.client.HTTP_PORT)
        self._password = password
        self._ctx = ctx
        self._ctx.ip = ip
        self._user = user

        if not self._ctx.auth_token:
            logger1.warning('Proceeding with authentication')
            self._authenticate()

    def _authenticate(self):
        http_body = {
            'aaaUser' : {
                'attributes': {
                    'name': self._user,
                    'pwd': self._password
                }
            }
        }

        http_response_code, http_response_header, http_response_payload = self._send_post("/api/aaaLogin.json?gui-token-request=yes", http_body)

        if http_response_code == 200:
            logger1.warning('Authentication successful')
            ltree = ListOfAciTree(http_response_payload)
            aaaLogin = to_tree(ltree.search('aaaLogin'))
            self._ctx.urlToken = aaaLogin.urlToken
            self._ctx.token = aaaLogin.token
            self._ctx.Cookie =  http_response_header('Set-Cookie')

        else:
            raise AuthFailure('Failed to login to {0}'.format(self._ctx.ip))


    def _send_get(self, url):
        try:
            self.__conn.request('GET', url, '', self._ctx.http_header)
            http_response = self.__conn.getresponse()
            http_response_payload = json.loads(http_response.read().decode('utf-8'))

        except json.decoder.JSONDecodeError as error:
            logger1.critical("Respond payload not in JSON format")
            raise

        except TimeoutError as error:
            logger1.critical("Connection Timeout")
            raise

        except Exception as error:
            logger1.critical(error)
            raise

        logger1.info('{0} - {1}'.format(url, http_response.code))
        return http_response.code, http_response.getheader, http_response_payload


    def _send_post(self, url, payload):
        try:
            self.__conn.request('POST', url, json.dumps(payload), self._ctx.http_header)
            http_response = self.__conn.getresponse()
            http_response_payload = json.loads(http_response.read().decode('utf-8'))

        except json.decoder.JSONDecodeError as error:
            logger1.critical("Respond payload not in JSON format")
            raise

        except TimeoutError as error:
            logger1.critical("Connection Timeout")
            raise

        except Exception as error:
            logger1.critical(error)
            raise

        logger1.info('{0} - {1}'.format(url, http_response.code))
        return http_response.code, http_response.getheader, http_response_payload


    def disconnect(self):
        self.__conn.close()
        self._ctx.reset_token()


class AuthRefreshTread(threading.Thread):
    def __init__(self, ctx=api_ctx, *args, **kwargs):
        self._ctx = ctx
        super(AuthRefreshTread, self).__init__(*args, **kwargs)

    def run(self):
        while self._ctx.keep_running_thread:
            time.sleep(60)

            logger1.warning('Refreshing auth token')

            self._ctx.lock.acquire()
            #conn = http.client.HTTPConnection(self._ctx.ip, http.client.HTTP_PORT)
            conn = http.client.HTTPSConnection(self._ctx.ip, http.client.HTTPS_PORT)
            conn.request('GET', "/api/aaaRefresh.json?gui-token-request=yes", '', self._ctx.http_header)
            http_response = conn.getresponse()
            http_response_payload = json.loads(http_response.read().decode('utf-8'))

            if http_response.code == 200:
                logger1.error('Token successfully refreshed')

                ltree = ListOfAciTree(http_response_payload)
                aaaLogin = to_tree(ltree.search('aaaLogin'))
                #shared.urlToken = aaaLogin.urlToken
                self._ctx.token = aaaLogin.token
                self._ctx.Cookie = http_response.getheader('Set-Cookie')

            else:
                logger1.error('Failed to refresh token with HTTP respond code {0}'.format(http_response.code))
                self._ctx.reset_token()
                self._ctx.keep_running_thread = False

            self._ctx.lock.release()
            conn.close()

        logger1.error("The authentication loop was stopped")


class Api(HttpClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def lookup_by_class(self, cls, **kwarg) -> ListOfAciTree:
        base_url = "/api/node/class/"
        url = base_url + cls + ".json?"

        for key, value in kwarg.items():
            url += "&" + key.replace('_', '-') + '=' + value

        http_response_code, http_response_header, http_response_payload = self._send_get(url)
        return ListOfAciTree(http_response_payload)

    def lookup_by_dn(self, dn, **kwarg) -> AciTree:
        base_url = "/api/mo/"
        url = base_url + dn + ".json?"

        for key, value in kwarg.items():
            url += "&" + key.replace('_', '-') + '=' + value

        http_response_code, http_response_header, http_response_payload = self._send_get(url)
        return AciTree(http_response_payload)

    def lookup(self, **kwarg) -> AciSet:
        if 'cls' in kwarg.keys():
            base_url = "/api/node/class/"
            url = base_url + kwarg.pop('cls') + ".json?"

        elif 'dn' in kwarg.keys():
            base_url = "/api/mo/"
            url = base_url + kwarg.pop('dn') + ".json?"

        else:
            raise InvalidLookupRequest("Failed to lookup. Neither a class or dn was provided")

        for key, value in kwarg.items():
            url += "&" + key.replace('_', '-') + '=' + value

        http_response_code, http_response_header, http_response_payload = self._send_get(url)
        return AciSet(http_response_payload)

    def mqapi2(self, tool, **kwarg):
        base_url = "/mqapi2/"
        url = base_url + tool + ".json?"

        for key, value in kwarg.items():
            url += "&" + key.replace('_', '-') + '=' + value

        http_response_code, http_response_header, http_response_payload = self._send_get(url)
        return ListOfAciTree(http_response_payload)

class SubscriptionRefreshTread(threading.Thread):
    def __init__(self, ctx=api_ctx, *args, **kwargs):
        self._ctx=ctx
        super(SubscriptionRefreshTread, self).__init__(*args, **kwargs)

    def run(self):
        while self._ctx.keep_running_thread:
            time.sleep(60)

            self._ctx.lock.acquire()
            #conn = http.client.HTTPConnection(self._ctx.ip, http.client.HTTP_PORT)
            conn = http.client.HTTPSConnection(self._ctx.ip, http.client.HTTPS_PORT)

            for sub in self._ctx.subscription_list:
                conn.request('GET', '/api/subscriptionRefresh.json?id={0}'.format(sub), '', self._ctx.http_header)
                http_response = conn.getresponse()
                http_response_payload = json.loads(http_response.read().decode('utf-8'))

                if http_response.code == 200:
                    logger1.error('Sub {0} successfully refresh'.format(sub))

                else:
                    logger1.error('Sub {0} failed to be refresh'.format(sub))
                    self._ctx.reset_token()
                    self._ctx.keep_running_thread = False

            self._ctx.lock.release()
            conn.close()

        logger1.error("The subscription loop was stopped")


class EApi(Api):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._ctx.keep_running_thread = True
        self._aaa_refresh_tread = AuthRefreshTread(ctx=self._ctx, name='aaaRefresh')
        self._sub_refresh_thread = None
        self._stream_handler_thread = None
        self._aaa_refresh_tread.start()

    def lookup_by_class(self, *args, **kwarg) -> ListOfAciTree:
        ltree = super().lookup_by_class(*args, **kwarg)
        self._watch_subscription(ltree)

        return ltree

    def lookup_by_dn(self, *args, **kwarg) -> AciTree:
        tree = super().lookup_by_dn(*args, **kwarg)
        self._watch_subscription(tree)

        return tree

    def lookup(self, *args, **kwarg) -> AciSet:
        aci_set = super().lookup(*args, **kwarg)
        self._watch_subscription(aci_set)

        return aci_set

    def setup_stream(self, stream_handler_class, *args, **kwargs):
        self._stream_handler_thread = stream_handler_class(ctx=self._ctx, *args, **kwargs)
        self._stream_handler_thread.daemon = True
        self._stream_handler_thread.start()

        for c in range(5):
            if self._ctx.ws_status == "OPEN":
                break

            logger1.warning('Waiting for WebSocket to be opened...')
            time.sleep(1)

        if self._ctx.ws_status != "OPEN":
            raise WsOpenFailure('Failed to opened WebSocket within 5 sec')

    def _watch_subscription(self, aci_set : AciSet):

        if aci_set.subscription_id and not self._ctx.subscription_list:
            logger1.info('First subscription received {0}'.format(aci_set.subscription_id))
            self._ctx.subscription_list.append(aci_set.subscription_id)

            logger1.warning('Starting subscription loop')
            self._sub_refresh_thread = SubscriptionRefreshTread()
            self._sub_refresh_thread.start()

        if aci_set.subscription_id and aci_set.subscription_id not in self._ctx.subscription_list:
            logger1.info('New subscription id {0}'.format(aci_set.subscription_id))
            self._ctx.subscription_list.append(aci_set.subscription_id)

    def hold(self):
        self._aaa_refresh_tread.join()
        self._stream_handler_thread.kill_thread()








from .object_set import AciSet, AciTree, ListOfAciTree, to_tree, mpprint
from abc import ABC, abstractmethod
import threading
import websocket
import logging
import ssl
import json

logger1 = logging.getLogger(__name__)


class StreamHandler(ABC, threading.Thread):
    def __init__(self, ctx, *args, **kwargs):
        super(StreamHandler, self).__init__(*args, **kwargs)
        self._stream = None
        self._ctx=ctx

    def run(self):
        ws_opt = {
            "cert_reqs": ssl.CERT_NONE,
            "check_hostname": False
        }

        logger1.error('Starting websocket with url {0}'.format(self._ctx.ws_url))

        #websocket.enableTrace(True)
        self._stream = websocket.WebSocketApp(self._ctx.ws_url, header=self._ctx.http_header, on_message=self._handler, on_error=self._on_error, on_close=self._on_close, on_open=self._on_open)
        self._stream.run_forever(sslopt=ws_opt)

    def _on_error(self, error):
        logger1.critical(error)
        self._ctx.ws_status = "ERROR"

    def _on_close(self):
        logger1.warning('WebSocket has been closed')
        self._ctx.ws_status = "CLOSE"

    def _on_open(self):
        logger1.warning('WebSocket has been opened')
        self._ctx.ws_status = "OPEN"

    def kill_thread(self):
        logger1.warning("Closing the WebSocket...")
        self._stream.close()

    @abstractmethod
    def _handler(self, message):
        pass



class PrintStream(StreamHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _handler(self, msg):
        logger1.debug('Websocket has received a message')
        obj = AciTree(json.loads(msg))

        mpprint(obj)







from .exception import CastingFailure
import logging
import pprint
import json

logger1 = logging.getLogger(__name__)

class AciSet:
    pass

class AciTree(AciSet):
    pass

class ListOfAciTree(AciSet):
    pass

class AciSet:
    def __init__(self, payload):
        self.__dict__['_total_count'] = None
        self.__dict__['_imdata'] = None
        self.__dict__['_subscription_id'] = None

        if 'totalCount' in payload.keys():
            self.__dict__['_total_count'] = int(payload['totalCount'])

        if 'imdata' in payload.keys():
            self.__dict__['_imdata'] = payload['imdata']

        if 'subscriptionId' in payload.keys():
            self.__dict__['_subscription_id'] = payload['subscriptionId']

    def ltree(self) -> ListOfAciTree:
        return ListOfAciTree({'imdata': self.imdata, 'totalCount': self.total_count})

    def tree(self) -> AciTree:
        if len(self) > 1:
            raise CastingFailure("There is more than 1 tree in the object set")

        return AciTree({'imdata': self.imdata, 'totalCount': self.total_count})

    def search(self, cls = 'unspecified', **kwargs) -> ListOfAciTree:
        '''
        :param aci_class: the name of the class to be found in self._imdata
        :param kwargs: dict of attributes:value
        :return: the list of objects present in self._imdata of class aci_class with attributes matching kwargs. Returns an empty list of no match
        '''
        res = self.__search(self._imdata, '', cls, [], kwargs)
        if len(res) == 0:
            logger1.debug('No object {0} found in the tree'.format(cls))

        #return res
        return ListOfAciTree({'imdata' : res, 'totalCount': len(res)})

    def __search(self, tree, dn, cls, res, kwargs):
        '''
        :param tree: a tree of objects
        :param aci_class: the name of the class to be found in self._imdata
        :param res: a list of object found so far in tree
        :return: the list of objects present in tree of class aci_class with attributes matching kwargs. Returns an empty list of no match
        '''

        if isinstance(tree, list):
            for child in tree:
                self.__search(child, dn, cls, res, kwargs)

        elif isinstance(tree, dict):
            missing_dn = False

            for key in tree.keys():
                if 'attributes' in tree[key].keys() and 'dn' in tree[key]['attributes'].keys():
                    dn = tree[key]['attributes']['dn']

                elif 'attributes' in tree[key].keys() and 'rn' in tree[key]['attributes'].keys():
                    missing_dn = True
                    dn = "{0}/{1}".format(dn, tree[key]['attributes']['rn'])

                if key == cls or cls == 'unspecified':
                    if 'attributes' in tree[key].keys() and self.__match_attribute(tree[key]['attributes'], kwargs):
                        logger1.debug('Object {0} found: {1}'.format(key, tree[key]['attributes']))

                        if missing_dn:
                            tree[key]['attributes']['dn'] = dn

                        res.append(tree)

                elif key != cls:
                    if 'children' in tree[key].keys():
                        self.__search(tree[key]['children'], dn, cls, res, kwargs)

        return res

    def __match_attribute(self, obj, kwargs):
        '''
        :param obj: A dictionary
        :param kwargs: A second dictionary containing the attributes:value to be found in obj
        :return: Return true if all atributes in kwargs are found in obj. False otherwise
        '''

        result = True

        for attr, value in kwargs.items():
            if attr not in obj.keys() or obj[attr] != value:
                result = False

        return result

    @property
    def total_count(self) -> str:
        return self._total_count

    @property
    def imdata(self) -> list:
        return self._imdata

    @property
    def subscription_id(self) -> str:
        return self._subscription_id

    def __len__(self):
        return self.total_count


class ListOfAciTree(AciSet):
    def __init__(self, response):
        super().__init__(response)

    def __iter__(self) -> AciTree:
        for el in self.imdata:
            yield AciTree({'imdata' : [el]})

    def __getitem__(self, item) -> AciTree:
        return AciTree({'imdata': [self.imdata[item]], 'totalCount': 1})

    def __repr__(self):
        return json.dumps(self.imdata)

    def __str__(self):
        return json.dumps(self.imdata)


class AciTree(AciSet):
    def __init__(self, response):
        super().__init__(response)

    def __getattr__(self, attr):
        cls = list(self.imdata[0].keys())[0]
        return self.imdata[0][cls]['attributes'][attr]

    def __setattr__(self, key, value):
        cls = list(self.imdata[0].keys())[0]
        self.imdata[0][cls]['attributes'][key] = value

    def __repr__(self):
        return json.dumps(self.imdata[0])

    def __str__(self):
        return json.dumps(self.imdata[0])

    def update_cls(self, new_cls):
        old_cls = list(self.imdata[0].keys())[0]
        self.imdata[0] = {new_cls: self.imdata[0][old_cls]}

    def cls(self) -> str:
        return list(self.imdata[0].keys())[0]

    def children(self) -> ListOfAciTree:
        if self.total_count > 0:
            cls = list(self.imdata[0].keys())[0]

            if "children" in self.imdata[0][cls].keys():
                return ListOfAciTree({'imdata' : self.imdata[0][cls]['children'], 'tocal_count': len(self.imdata[0][cls]['children'])})

        else:
            return ListOfAciTree({'imdata': [], 'total_count': 0})


def to_tree(ltree: ListOfAciTree) -> AciTree:
    return AciSet.tree(ltree)

def to_ltree(tree: AciTree) -> ListOfAciTree:
    return AciSet.ltree(tree)

def mpprint(obj: AciSet):
    pprint.pprint(obj.imdata[0])

